"""
Handle input and output files for IceProd.

This module can be called standalone with:

.. code-block:: bash

    python -m iceprod.core.io -f CONFIG_FILE -d DIRECTORY {input,output}
"""

import os
import asyncio
import logging

from iceprod.core.serialization import serialize_json
from iceprod.core.exe import Config, SetupEnv

logger = logging.getLogger('io')


def get_current_task(config):
    name = config['options']['task']
    for task in config['tasks']:
        if task['name'] == name:
            return task
    raise Exception('cannot find task in config')


async def process(config):
    task = get_current_task(config)
    cfg = Config(config=config, logger=logger)
    env_opts = cfg.parseObject(config['options'], {})
    async with SetupEnv(cfg, config['steering'], {'options':env_opts}, logger=logger) as env:
        async with SetupEnv(cfg, task, env, logger=logger) as env2:
            for tray in task['trays']:
                for i in range(tray['iterations']):
                    cfg.config['options']['iter'] = i
                    async with SetupEnv(cfg, tray, env2, logger=logger) as env3:
                        for module in tray['modules']:
                            async with SetupEnv(cfg, module, env3, logger=logger) as env:
                                pass


async def handle_input(config):
    """Handle downloading all input files for specified task"""
    # first, filter out uploaded files
    def process_obj(obj):
        if 'data' in obj:
            new_data = []
            for data in obj['data']:
                if data['movement'] in ('input','both'):
                    new_data.append(data)
            obj['data'] = new_data
    process_obj(config['steering'])
    for task in config['tasks']:
        process_obj(task)
        for tray in task['trays']:
            process_obj(tray)
            for module in tray['modules']:
                process_obj(module)

    # now do the download
    await process(config)


async def handle_output(config):
    """Handle uploading all output files for specified task"""
    # first, filter out downloaded files
    def process_obj(obj):
        if 'data' in obj:
            new_data = []
            for data in obj['data']:
                if data['movement'] in ('output','both'):
                    new_data.append(data)
            obj['data'] = new_data
        if 'resources' in obj:
            obj['resources'] = []
        if 'classes' in obj:
            obj['classes'] = []
    process_obj(config['steering'])
    for task in config['tasks']:
        process_obj(task)
        for tray in task['trays']:
            process_obj(tray)
            for module in tray['modules']:
                process_obj(module)

    # now do the upload
    await process(config)


def setup_directory(config, directory):
    """Set up directory in config"""
    os.chdir(directory)

    # make sure some basic options are set
    if 'resource_url' not in config['options']:
        config['options']['resource_url'] = 'http://prod-exe.icecube.wisc.edu/'
    if 'offline' not in config['options']:
        config['options']['offline'] = True
    if 'offline_transfer' not in config['options']:
        config['options']['offline_transfer'] = True
    if 'data_url' not in config['options']:
        config['options']['data_url'] = 'gsiftp://gridftp.icecube.wisc.edu/'
    if 'svn_repository' not in config['options']:
        config['options']['svn_repository'] = 'http://code.icecube.wisc.edu/svn/'
    if 'site_temp' not in config['options']:
        config['options']['site_temp'] = 'gsiftp://gridftp-scratch.icecube.wisc.edu/local/simprod/'
    if 'dataset_temp' not in config['options']:
        config['options']['dataset_temp'] = os.path.join(config['options']['site_temp'],'$(dataset)')
    if 'job_temp' not in config['options']:
        config['options']['job_temp'] = os.path.join(config['options']['dataset_temp'],'$(job)')
    if 'subprocess_dir' not in config['options']:
        config['options']['subprocess_dir'] = directory
    if 'task_temp' not in config['options']:
        config['options']['task_temp'] = os.path.join(config['options']['subprocess_dir'],'task_temp')
    if 'tray_temp' not in config['options']:
        config['options']['tray_temp'] = os.path.join(config['options']['subprocess_dir'],'tray_temp')
    if 'local_temp' not in config['options']:
        config['options']['local_temp'] = os.path.join(config['options']['subprocess_dir'],'local_temp')


def main():
    import argparse
    parser = argparse.ArgumentParser(description='Handle IceProd I/O')
    parser.add_argument('-f','--config_file',help='IceProd json config file')
    parser.add_argument('-d','--directory',help='Directory to read from / write to')
    parser.add_argument('transfer', choices=['input','output'], help='type of transfer')
    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO)

    config = serialize_json.load(args.config_file)

    setup_directory(config, args.directory)
    if args.transfer == 'input':
        fut = handle_input(config)
    elif args.transfer == 'output':
        fut = handle_output(config)
    else:
        raise Exception('bad transfer type')

    loop = asyncio.get_event_loop()
    loop.run_until_complete(fut)


if __name__ == '__main__':
    main()



"""
A set of classes that holds dataset configuration data.

Each class is based on a dictionary and only contains simple elements for
easy serialization to json (or other formats). A full dataset configuration
can be had by serializing the :class:`Job` class.

The `convert` method of each class will turn any regular dictionary objects
into special `dataclasses` objects.

The `valid` method of each class will test the validity of the data to be
an actual dataset config.

The `output` method of each class will create json with info on each
dataclass, to be used in javascript.
"""

from __future__ import absolute_import, division, print_function

import time

from numbers import Number, Integral
try:
    String = basestring
except NameError:
    String = str

# pluralizations for keys that are not classes here
_plurals = {
    'Option': 'Options',
    'Category': 'Categories',
    'Parameter': 'Parameters',
    'System': 'System',
    'Dependence': 'Depends'
}


class Job(dict):
    """
    Holds all information about a running job.

    If the `options` are empty, this is the same as a dataset
    configuration.

    :ivar version: 3
    :ivar options: {} -- a dict of parameters to pass to the task runner
    :ivar steering: None
    :ivar tasks: []
    :ivar difplus: None
    :ivar description: ""
    :ivar categories: []
    """
    plural = 'Jobs'

    def __init__(self,*args,**kwargs):
        self['version'] = 3
        self['options'] = {}
        self['steering'] = None
        self['tasks'] = []
        self['difplus'] = None
        self['description'] = ''
        self['categories'] = []
        super(Job,self).__init__(*args,**kwargs)

    def output(self):
        """Output dict with values and (optionally) the object name for
        new objects."""
        ret = {}
        for n in self:
            if n == 'tasks':
                ret[n] = [self[n],'Task']
            elif n == 'difplus':
                ret[n] = [self[n],'DifPlus']
            elif n == 'steering':
                ret[n] = [self[n],'Steering']
            elif n == 'categories':
                ret[n] = [self[n],'']
            else:
                ret[n] = self[n]
        return ret

    def convert(self):
        if (self['steering'] is not None and
                not isinstance(self['steering'],Steering)):
            tmp = Steering(self['steering'])
            tmp.convert()
            self['steering'] = tmp
        for i,t in enumerate(self['tasks']):
            if not isinstance(t,Task):
                tmp = Task(t)
                tmp.convert()
                self['tasks'][i] = tmp
        if (self['difplus'] is not None and
                not isinstance(self['difplus'],DifPlus)):
            tmp = DifPlus(self['difplus'])
            tmp.convert()
            self['difplus'] = tmp

    def valid(self):
        try:
            return (
                isinstance(self['version'],Number)
                and self['version'] >= 3
                and isinstance(self['options'],dict)
                and (self['steering'] is None or (
                    isinstance(self['steering'],Steering)
                    and self['steering'].valid()))
                and isinstance(self['tasks'],list)
                and all(isinstance(t,Task) and t.valid() for t in self['tasks'])
                and (self['difplus'] is None or (
                    isinstance(self['difplus'],DifPlus)
                    and self['difplus'].valid()))
                and isinstance(self['description'],String)
                and isinstance(self['categories'],list)
            )
        except Exception:
            return False


class Steering(dict):
    """
    Holds all information that goes in the steering section of a configuration.

    :ivar parameters: {}
    :ivar batchsys: None
    :ivar system: {} -- just specialized parameters
    :ivar resources: []
    :ivar data: []
    """
    plural = 'Steering'

    def __init__(self,*args,**kwargs):
        self['parameters'] = {}
        self['batchsys'] = None
        self['system'] = {}
        self['resources'] = []
        self['data'] = []
        super(Steering,self).__init__(*args,**kwargs)

    def output(self):
        """Output dict with values and (optionally) the object name for
        new objects."""
        ret = {}
        for n in self:
            if n == 'resources':
                ret[n] = [self[n],'Resource']
            elif n == 'data':
                ret[n] = [self[n],'Data']
            elif n == 'batchsys':
                ret[n] = [self[n],'Batchsys']
            else:
                ret[n] = self[n]
        return ret

    def convert(self):
        if (self['batchsys'] is not None and
                not isinstance(self['batchsys'],Batchsys)):
            tmp = Batchsys(self['batchsys'])
            tmp.convert()
            self['batchsys'] = tmp
        for i,r in enumerate(self['resources']):
            if not isinstance(r,Resource):
                tmp = Resource(r)
                tmp.convert()
                self['resources'][i] = tmp
        for i,d in enumerate(self['data']):
            if not isinstance(d,Data):
                tmp = Data(d)
                tmp.convert()
                self['data'][i] = tmp

    def valid(self):
        try:
            return (
                isinstance(self['parameters'],dict)
                and (self['batchsys'] is None or (
                    isinstance(self['batchsys'],Batchsys)
                    and self['batchsys'].valid()))
                and isinstance(self['system'],dict)
                and isinstance(self['resources'],list)
                and all(isinstance(r,Resource) and r.valid() for r in self['resources'])
                and isinstance(self['data'],list)
                and all(isinstance(d,Data) and d.valid() for d in self['data'])
            )
        except Exception:
            return False


class Batchsys(dict):
    """
    Holds information for running on grid/cluster types.
    Designed as a dict of dicts, one for each grid/cluster type.
    """
    plural = "Batchsys"

    def output(self):
        """Output dict with values and (optionally) the object name for
        new objects."""
        return {'*':{}}

    def convert(self):
        pass

    def valid(self):
        try:
            return all(isinstance(v,dict) for v in self.values())
        except Exception:
            return False


class _TaskCommon(dict):
    """
    Holds common attributes used by task, tray, module.

    :ivar name: ''
    :ivar resources: []
    :ivar data: []
    :ivar classes: []
    :ivar parameters: {}
    """
    def __init__(self,*args,**kwargs):
        self['name'] = ''
        self['resources'] = []
        self['data'] = []
        self['classes'] = []
        self['parameters'] = {}
        super(_TaskCommon,self).__init__(*args,**kwargs)

    def convert(self):
        for i,r in enumerate(self['resources']):
            if not isinstance(r,Resource):
                tmp = Resource(r)
                tmp.convert()
                self['resources'][i] = tmp
        for i,d in enumerate(self['data']):
            if not isinstance(d,Data):
                tmp = Data(d)
                tmp.convert()
                self['data'][i] = tmp
        for i,c in enumerate(self['classes']):
            if not isinstance(c,Class):
                tmp = Class(c)
                tmp.convert()
                self['classes'][i] = tmp

    def valid(self):
        try:
            return (
                isinstance(self['name'],String)
                and isinstance(self['resources'],list)
                and all(isinstance(r,Resource) and r.valid() for r in self['resources'])
                and isinstance(self['data'],list)
                and all(isinstance(d,Data) and d.valid() for d in self['data'])
                and isinstance(self['classes'],list)
                and all(isinstance(c,Class) and c.valid() for c in self['classes'])
                and isinstance(self['parameters'],dict)
            )
        except Exception:
            return False


class Requirement(dict):
    plural = 'Requirements'

    def __init__(self, *args,**kwargs):
        self['cpu'] = None
        self['gpu'] = None
        self['memory'] = None
        self['disk'] = None
        self['os'] = None
        self['site'] = None
        super(Requirement,self).__init__(*args,**kwargs)

    def output(self):
        """Output dict with values and (optionally) the object name for
        new objects."""
        ret = {}
        for n in self:
            ret[n] = self[n]
        return ret

    def convert(self):
        pass

    def valid(self):
        return (
            (self['cpu'] is None or isinstance(self['cpu'],(String,Number)))
            and (self['gpu'] is None or isinstance(self['gpu'],(String,Number)))
            and (self['memory'] is None or isinstance(self['memory'],(String,Number)))
            and (self['disk'] is None or isinstance(self['disk'],(String,Number)))
            and (self['os'] is None or isinstance(self['os'],String))
            and (self['site'] is None or isinstance(self['site'],String))
        )


class Task(_TaskCommon):
    """
    Holds all information about a task.

    :ivar depends: [] -- a list of task names
    :ivar batchsys: None
    :ivar trays: []
    :ivar requirements: {} -- a dict of requirements
    :ivar task_files: False -- whether to use the task files API
    """
    plural = 'Tasks'

    def __init__(self,*args,**kwargs):
        self['depends'] = []
        self['batchsys'] = None
        self['trays'] = []
        self['requirements'] = Requirement()
        self['task_files'] = False
        super(Task,self).__init__(*args,**kwargs)

    def output(self):
        """Output dict with values and (optionally) the object name for
        new objects."""
        ret = {}
        for n in self:
            if n == 'resources':
                ret[n] = [self[n],'Resource']
            elif n == 'data':
                ret[n] = [self[n],'Data']
            elif n == 'classes':
                ret[n] = [self[n],'Class']
            elif n == 'depends':
                ret[n] = [self[n],'']
            elif n == 'batchsys':
                ret[n] = [self[n],'Batchsys']
            elif n == 'trays':
                ret[n] = [self[n],'Tray']
            elif n == 'requirements':
                ret[n] = [self[n],'Requirement']
            else:
                ret[n] = self[n]
        return ret

    def convert(self):
        super(Task,self).convert()
        if (self['batchsys'] is not None and
                not isinstance(self['batchsys'],Batchsys)):
            tmp = Batchsys(self['batchsys'])
            tmp.convert()
            self['batchsys'] = tmp
        for i,t in enumerate(self['trays']):
            if not isinstance(t,Tray):
                tmp = Tray(t)
                tmp.convert()
                self['trays'][i] = tmp

    def valid(self):
        try:
            return (
                super(Task,self).valid()
                and isinstance(self['depends'],list)
                and all(isinstance(r,(String,Number)) for r in self['depends'])
                and (self['batchsys'] is None or (
                    isinstance(self['batchsys'],Batchsys)
                    and self['batchsys'].valid()))
                and isinstance(self['trays'],list)
                and all(isinstance(t,Tray) and t.valid() for t in self['trays'])
                and isinstance(self['requirements'],Requirement)
                and self['requirements'].valid()
                and isinstance(self['task_files'],bool)
            )
        except Exception:
            return False


class Tray(_TaskCommon):
    """
    Holds all information about a tray.

    :ivar iterations: 1
    :ivar modules: []
    """
    plural = 'Trays'

    def __init__(self,*args,**kwargs):
        self['iterations'] = 1
        self['modules'] = []
        super(Tray,self).__init__(*args,**kwargs)

    def output(self):
        """Output dict with values and (optionally) the object name for
        new objects."""
        ret = {}
        for n in self:
            if n == 'resources':
                ret[n] = [self[n],'Resource']
            elif n == 'data':
                ret[n] = [self[n],'Data']
            elif n == 'classes':
                ret[n] = [self[n],'Class']
            elif n == 'modules':
                ret[n] = [self[n],'Module']
            else:
                ret[n] = self[n]
        return ret

    def convert(self):
        super(Tray,self).convert()
        for i,m in enumerate(self['modules']):
            if not isinstance(m,Module):
                tmp = Module(m)
                tmp.convert()
                self['modules'][i] = tmp

    def valid(self):
        try:
            return (
                super(Tray,self).valid()
                and isinstance(self['iterations'],Integral)
                and isinstance(self['modules'],list)
                and all(isinstance(m,Module) and m.valid() for m in self['modules'])
            )
        except Exception:
            return False


class Module(_TaskCommon):
    """
    Holds all information about a module.

    :ivar running_class: None -- the python class or function to call
    :ivar src: None -- src of class or script
    :ivar args: None -- args to give to class or src if not an iceprod module
    :ivar env_shell: None -- src of script which sets env and calls arg
    :ivar env_clear: True -- clear the env before calling the module
                              (calls env_shell after clearing, if defined)
    :ivar configs: None -- any json config files that should be written
                             (format is {filename: data})

    Note that `env_clear` should be used carefully, as it wipes out
    any loaded classes.
    """
    plural = 'Modules'

    def __init__(self,*args,**kwargs):
        self['running_class'] = ''
        self['src'] = ''
        self['args'] = ''
        self['env_shell'] = ''
        self['env_clear'] = True
        self['configs'] = {}
        super(Module,self).__init__(*args,**kwargs)

    def output(self):
        """Output dict with values and (optionally) the object name for
        new objects."""
        ret = {}
        for n in self:
            if n == 'resources':
                ret[n] = [self[n],'Resource']
            elif n == 'data':
                ret[n] = [self[n],'Data']
            elif n == 'classes':
                ret[n] = [self[n],'Class']
            else:
                ret[n] = self[n]
        return ret

    def convert(self):
        super(Module,self).convert()

    def valid(self):
        try:
            return (
                super(Module,self).valid()
                and isinstance(self['running_class'],String)
                and isinstance(self['src'],String)
                and isinstance(self['env_shell'],String)
                and isinstance(self['env_clear'],bool)
                and isinstance(self['configs'],dict)
            )
        except Exception:
            return False


class Class(dict):
    """
    A class object, downloaded from a url.

    :ivar name: None -- required
    :ivar src: None -- if downloaded from url
    :ivar resource_name: None -- if present in resource object
    :ivar recursive: False
    :ivar libs: None -- if more than default lib directory
    :ivar env_vars: None
    """
    plural = 'Classes'

    def __init__(self,*args,**kwargs):
        self['name'] = ''
        self['src'] = ''
        self['resource_name'] = ''
        self['recursive'] = False
        self['libs'] = ''
        self['env_vars'] = ''
        super(Class,self).__init__(*args,**kwargs)

    def output(self):
        """Output dict with values and (optionally) the object name for
        new objects."""
        ret = {}
        for n in self:
            ret[n] = self[n]
        return ret

    def convert(self):
        pass

    def valid(self):
        try:
            return (
                isinstance(self['name'],String)
                and isinstance(self['src'],String)
                and isinstance(self['resource_name'],String)
                and isinstance(self['recursive'],bool)
                and isinstance(self['libs'],String)
                and isinstance(self['env_vars'],String)
            )
        except Exception:
            return False


class _ResourceCommon(dict):
    """
    Holds common attributes used by Resource and Data.

    :ivar remote: ''
    :ivar local: ''
    :ivar compression: False
    """
    compression_options = [False,True,'none','gzip','gz','bzip','bz2','lzma']

    def __init__(self,*args,**kwargs):
        self['remote'] = ''
        self['local'] = ''
        self['compression'] = False
        self['transfer'] = True
        super(_ResourceCommon,self).__init__(*args,**kwargs)

    def convert(self):
        pass

    def valid(self):
        try:
            return (
                isinstance(self['remote'],String)
                and isinstance(self['local'],String)
                and self['compression'] in self.compression_options
                and isinstance(self['transfer'],(String,Number,bool))
            )
        except Exception:
            return False

    def do_transfer(self):
        """
        Test if we should actually transfer the file.
        """
        ret = True
        if isinstance(self['transfer'], bool):
            ret = self['transfer']
        elif isinstance(self['transfer'], String):
            t = self['transfer'].lower()
            if t in ('n','no','not','f','false'):
                ret = False
            elif t in ('m','maybe','exists') or t.startswith('if'):
                ret = 'maybe'
        elif isinstance(self['transfer'], Number):
            if self['transfer'] == 0:
                ret = False
        return ret


class Resource(_ResourceCommon):
    """
    A resource object, representing a file to download.

    :ivar arch: None
    """
    plural = 'Resources'

    def __init__(self,*args,**kwargs):
        self['arch'] = ''
        super(Resource,self).__init__(*args,**kwargs)

    def output(self):
        """Output dict with values and (optionally) the object name for
        new objects."""
        ret = {}
        for n in self:
            if n == 'compression':
                ret[n] = [self[n],self.compression_options]
            ret[n] = self[n]
        return ret

    def convert(self):
        super(Resource,self).convert()

    def valid(self):
        try:
            return (
                super(Resource,self).valid()
                and isinstance(self['arch'],String)
            )
        except Exception:
            return False


class Data(_ResourceCommon):
    """
    A data object, representing input and/or output of data.

    :ivar type: 'permanent' -- required
    :ivar movement: 'both' -- required
    """
    plural = 'Data'
    type_options = ['permanent','tray_temp','task_temp','job_temp','dataset_temp','site_temp']
    movement_options = ['input','output','both']

    def __init__(self,*args,**kwargs):
        self['type'] = 'permanent'
        self['movement'] = 'both'
        super(Data,self).__init__(*args,**kwargs)

    def output(self):
        """Output dict with values and (optionally) the object name for
        new objects."""
        ret = {}
        for n in self:
            if n == 'compression':
                ret[n] = [self[n],self.compression_options]
            elif n == 'type':
                ret[n] = [self[n],self.type_options]
            elif n == 'movement':
                ret[n] = [self[n],self.movement_options]
            else:
                ret[n] = self[n]
        return ret

    def convert(self):
        super(Data,self).convert()

    def valid(self):
        try:
            return (
                super(Data,self).valid()
                and self['type'] in self.type_options
                and self['movement'] in self.movement_options
            )
        except Exception:
            return False

    def storage_location(self,env):
        """
        Get storage location from the environment.

        :param env: environment
        :returns: storage location as a string, or raises Exception
        """
        type = self['type'].lower()
        if type not in Data.type_options:
            raise Exception('Data.type is undefined')
        if 'options' in env and type in env['options']:
            return env['options'][type]
        elif type == 'permanent':
            if 'options' in env and 'data_url' in env['options']:
                return env['options']['data_url']
            else:
                raise Exception('data_url not defined in env[\'options\']')
        else:
            raise Exception('%s not defined in env' % type)

    def filecatalog(self,env):
        """
        Get filecatalog from the environment.

        :param env: environment
        :returns: FileCatalog object, or raises Exception
        """
        type = self['type'].lower()
        if type not in Data.type_options:
            raise Exception('Data.type is undefined')
        elif type in ('job_temp','dataset_temp','site_temp'):
            if 'options' in env and 'filecatalog_temp' in env['options']:
                return env['options']['filecatalog_temp']
        elif type == 'permanent':
            if 'options' in env and 'filecatalog' in env['options']:
                return env['options']['filecatalog']
        raise Exception('%s not defined in env' % type)


class DifPlus(dict):
    """
    A DifPlus object.

    :ivar dif: None
    :ivar plus: None
    """
    plural = 'DifPlus'

    def __init__(self,*args,**kwargs):
        self['dif'] = None
        self['plus'] = None
        super(DifPlus,self).__init__(*args,**kwargs)

    def output(self):
        """Output dict with values and (optionally) the object name for
        new objects."""
        ret = {}
        for n in self:
            if n == 'dif':
                ret[n] = [self[n],'Dif']
            elif n == 'plus':
                ret[n] = [self[n],'Plus']
            else:
                ret[n] = self[n]
        return ret

    def convert(self):
        if self['dif'] and not isinstance(self['dif'],Dif):
            tmp = Dif(self['dif'])
            tmp.convert()
            self['dif'] = tmp
        if self['plus'] and not isinstance(Plus):
            tmp = Plus(self['plus'])
            tmp.convert()
            self['plus'] = tmp

    def valid(self):
        try:
            return (
                (self['dif'] is None or (isinstance(self['dif'],Dif) and self['dif'].valid()))
                and (self['plus'] is None or (isinstance(self['plus'],Plus) and self['plus'].valid()))
            )
        except Exception:
            return False


class Dif(dict):
    """
    A Dif object.

   :ivar entry_id: None
   :ivar entry_title: None
   :ivar parameters: ' '
   :ivar iso_topic_category: 'geoscientificinformation'
   :ivar data_ceter: None
   :ivar summary: ' '
   :ivar metadata_name: '[CEOS IDN DIF]'
   :ivar metadata_version: '9.4'
   :ivar personnel: []
   :ivar sensor_name: 'ICECUBE'
   :ivar source_name: 'SIMULATION'
   :ivar dif_creation_date: time.strftime("%Y-%m-%d")
    """
    plural = 'Dif'
    # TODO: move these to the DB, or somewhere IceCube-specific
    valid_parameters = [
        "SPACE SCIENCE > Astrophysics > Neutrinos",
        "SPACE SCIENCE > Astrophysics > Neutrinos > Atmospheric",
        "SPACE SCIENCE > Astrophysics > Neutrinos > Extraterrestrial Point Source",
        "SPACE SCIENCE > Astrophysics > Neutrinos > Gamma Ray Burst",
        "SPACE SCIENCE > Astrophysics > Neutrinos > WIMPS",
        "SPACE SCIENCE > Astrophysics > Neutrinos > Diffuse Source",
        "SPACE SCIENCE > Astrophysics > Neutrinos > Extreme High Energy",
        "SPACE SCIENCE > Astrophysics > Neutrinos > Super Nova",
        "SPACE SCIENCE > Astrophysics > Neutrinos > Cosmic Ray Muon Component",
        "SPACE SCIENCE > Astrophysics > Neutrinos > Tau",
        "SPACE SCIENCE > Astrophysics > Neutrinos > Cascades",
        "SPACE SCIENCE > Astrophysics > Neutrinos > Galactic Plane",
        "SPACE SCIENCE > Astrophysics > Cosmic Rays",
        "SPACE SCIENCE > Astrophysics > Cosmic Rays > Composition",
        "SPACE SCIENCE > Astrophysics > Cosmic Rays > Air Shower",
        "SPACE SCIENCE > Astrophysics > Cosmic Rays > Cosmic Ray Muons",
        "SPACE SCIENCE > Astrophysics > Cosmic Rays > Moon Shadow",
        "SPACE SCIENCE > Engineering > Sensor Characteristics",
        "SPACE SCIENCE > Engineering > Sensor Characteristics > Photomultiplier Tubes",
        "SPACE SCIENCE > Engineering > Sensor Characteristics > Digital Optical Modules",
        "EARTH SCIENCE > Cryosphere > Glaciers/Ice Sheets",
        "EARTH SCIENCE > Cryosphere > Glaciers/Ice Sheets > Hot Water Drilling",
        "EARTH SCIENCE > Cryosphere > Glaciers/Ice Sheets > Hot Water Drilling > Hole Drilling",
        "EARTH SCIENCE > Cryosphere > Glaciers/Ice Sheets > Hot Water Drilling > Hole Refreeze"
    ]
    valid_source_name = {
        "SIMULATION":"Data which are numerically generated",
        "EXPERIMENTAL":"Data with an instrumentation based source"
    }
    valid_sensor_name = {
        "AMANDA-A":"Prototype Antarctic Muon and Neutrino Detector Array",
        "AMANDA":"Antarctic Muon and Neutrino Detector Array",
        "SPASE-1":"South Pole Air Shower Experiment 1",
        "SPASE-2":"South Pole Air Shower Experiment 2",
        "VULCAN":"South Pole Air-Cherenkov Telescope",
        "RICE":"Radio Ice Cherenkov Experiment",
        "GASP":"Gamma Astronomy-South Pole",
        "ICECUBE":"IceCube",
        "ICETOP":"IceTop",
        "EHWD":"Enhanced Hot Water Drill",
        "SPTR":"South Pole TDRSS Relay",
        "RPSC-MET":"Raytheon Polar Services Corporation Meteorology"
    }

    def __init__(self,*args,**kwargs):
        self['entry_id'] = None
        self['entry_title'] = None
        self['parameters'] = ''
        self['iso_topic_category'] = 'geoscientificinformation'
        self['data_ceter'] = None
        self['summary'] = ''
        self['metadata_name'] = '[CEOS IDN DIF]'
        self['metadata_version'] = '9.4'
        self['personnel'] = []
        self['sensor_name'] = 'ICECUBE'
        self['source_name'] = 'SIMULATION'
        self['dif_creation_date'] = time.strftime("%Y-%m-%d")
        super(Dif,self).__init__(*args,**kwargs)

    def output(self):
        """Output dict with values and (optionally) the object name for
        new objects."""
        ret = {}
        for n in self:
            if n == 'personnel':
                ret[n] = [self[n],'Personnel']
            else:
                ret[n] = self[n]
        return ret

    def convert(self):
        for i,p in enumerate(self['personnel']):
            if not isinstance(p,Personnel):
                tmp = Personnel(p)
                tmp.convert()
                self['personnel'][i] = tmp

    def valid(self):
        try:
            return (
                (self['entry_id'] is None or isinstance(self['entry_id'],(Number,String)))
                and (self['entry_title'] is None or isinstance(self['entry_title'],String))
                and isinstance(self['parameters'],String)
                and isinstance(self['iso_topic_category'],String)
                and (self['data_ceter'] is None or (
                    isinstance(self['data_ceter'],DataCenter)
                    and self['data_center'].valid())
                )
                and isinstance(self['summary'],String)
                and isinstance(self['metadata_name'],String)
                and isinstance(self['metadata_version'],(Number,String))
                and all(isinstance(p,Personnel) and p.valid() for p in self['personnel'])
                and isinstance(self['sensor_name'],String)
                and isinstance(self['source_name'],String)
                and isinstance(self['dif_creation_date'],(Number,String))
            )
        except Exception:
            return False


class Plus(dict):
    """
    A Plus object.

   :ivar start: None
   :ivar end: None
   :ivar category: None
   :ivar subcategory: None
   :ivar run_number: None
   :ivar i3db_key: None
   :ivar simdb_key: None
   :ivar project: [] -- [{name: version}]
   :ivar steering_file: None
   :ivar log_file: None
   :ivar command_line: None
    """
    plural = 'Plus'

    valid_category = [
        "unclassified",
        "generated",
        "propagated",
        "unbiased",
        "filtered",
        "calibration",
        "monitoring",
        "webcam",
        "hole",
        "TestDAQ",
        "FAT",
        "log",
        "upmu",
        "minbias",
        "cascades",
        "high-energy",
        "wimp",
        "GRB"
    ]

    def __init__(self,*args,**kwargs):
        self['start'] = None
        self['end'] = None
        self['category'] = None
        self['subcategory'] = None
        self['run_number'] = None
        self['i3db_key'] = None
        self['simdb_key'] = None
        self['project'] = []
        self['steering_file'] = None
        self['log_file'] = None
        self['command_line'] = None
        super(Plus,self).__init__(*args,**kwargs)

    def output(self):
        """Output dict with values and (optionally) the object name for
        new objects."""
        ret = {}
        for n in self:
            if n == 'project':
                ret[n] = [self[n],{}]
            else:
                ret[n] = self[n]
        return ret

    def convert(self):
        pass

    def valid(self):
        try:
            return (
                (self['start'] is None or isinstance(self['start'],(Number,String)))
                and (self['end'] is None or isinstance(self['end'],(Number,String)))
                and (self['category'] is None or isinstance(self['category'],String))
                and (self['subcategory'] is None or isinstance(self['subcategory'],String))
                and (self['run_number'] is None or isinstance(self['run_number'],(Number,String)))
                and (self['i3db_key'] is None or isinstance(self['i3db_key'],(Number,String)))
                and (self['simdb_key'] is None or isinstance(self['simdb_key'],(Number,String)))
                and all((
                    isinstance(p,dict)
                    and all(isinstance(k,String) for k in p.keys())
                    and all(isinstance(v,(Number,String)) for v in p.values())
                ) for p in self['project'])
                and (self['steering_file'] is None or isinstance(self['steering_file'],String))
                and (self['log_file'] is None or isinstance(self['log_file'],String))
                and (self['command_line'] is None or isinstance(self['command_line'],String))
            )
        except Exception:
            return False


class Personnel(dict):
    """
    A Personnel object.

   :ivar role: None
   :ivar first_name: None
   :ivar last_name: None
   :ivar email: None
    """
    plural = 'Personnel'

    def __init__(self,*args,**kwargs):
        self['role'] = None
        self['first_name'] = None
        self['last_name'] = None
        self['email'] = None
        super(Personnel,self).__init__(*args,**kwargs)

    def output(self):
        """Output dict with values and (optionally) the object name for
        new objects."""
        ret = {}
        for n in self:
            ret[n] = self[n]
        return ret

    def convert(self):
        pass

    def valid(self):
        try:
            return (
                (self['role'] is None or isinstance(self['role'],String))
                and (self['first_name'] is None or isinstance(self['first_name'],String))
                and (self['last_name'] is None or isinstance(self['last_name'],String))
                and (self['email'] is None or isinstance(self['email'],String))
            )
        except Exception:
            return False


class DataCenter(dict):
    """
    A Data Center object.

   :ivar name: None
   :ivar personnel: []
    """
    plural = 'DataCenter'
    valid_names = ['UWI-MAD/A3RI > Antarctic Astronomy and Astrophysics Research Institute, University of Wisconsin, Madison']

    def __init__(self,*args,**kwargs):
        self['name'] = None
        self['personnel'] = []
        super(DataCenter,self).__init__(*args,**kwargs)

    def output(self):
        """Output dict with values and (optionally) the object name for
        new objects."""
        ret = {}
        for n in self:
            if n == 'personnel':
                ret[n] = [self[n],'Personnel']
            else:
                ret[n] = self[n]
        return ret

    def convert(self):
        for i,p in enumerate(self['personnel']):
            if not isinstance(p,Personnel):
                tmp = Personnel(p)
                tmp.convert()
                self['personnel'][i] = tmp

    def valid(self):
        try:
            return (
                (self['name'] is None or isinstance(self['name'],String))
                and all(isinstance(p,Personnel) and p.valid() for p in self['personnel'])
            )
        except Exception:
            return False



"""
Detailed configuration for IceProd
"""

from __future__ import absolute_import, division, print_function

import os
import logging

from iceprod.core.jsonUtil import json_encode, json_decode
from iceprod.server import GlobalID, get_pkgdata_filename

import json
try:
    from jsonschema import validate
    from jsonschema.exceptions import ValidationError
except ImportError:
    validate = None
    ValidationError = Exception

logger = logging.getLogger('config')


def locateconfig(filename):
    """Locate a config file"""
    cfgpaths = [os.path.expandvars('$I3PROD')]
    if os.getcwd() not in cfgpaths:
        cfgpaths.append(os.getcwd())
    cfgpath = get_pkgdata_filename('iceprod.server','data')
    if cfgpath:
        cfgpaths.append(cfgpath)
    for cfgpath in list(cfgpaths):
        # try for an etc directory
        i = cfgpaths.index(cfgpath)
        if os.path.isdir(os.path.join(cfgpath,'etc')):
            cfgpaths.insert(i,os.path.join(cfgpath,'etc'))
            # try for an iceprod directory
            if os.path.isdir(os.path.join(cfgpath,'etc','iceprod')):
                cfgpaths.insert(i,os.path.join(cfgpath,'etc','iceprod'))
    for cfgpath in cfgpaths:
        if os.path.isfile(os.path.join(cfgpath,filename)):
            return os.path.join(cfgpath,filename)
    raise Exception('config {} not found'.format(filename))


class IceProdConfig(dict):
    """
    IceProd configuration.

    The main iceprod configuration. Designed to be modified in-program,
    not worrying about hand-editing. Currently uses a json file as backing.

    Use just like a dictionary. Note that load() and save() are called
    automatically, but are available for manual calling.

    Note that this class is not thread-safe.

    :param filename: filename for config file (optional)
    :param defaults: use default values (optional: default True)
    :param validate: turn validation on/off (optional: default True)
    """
    def __init__(self, filename=None, defaults=True, validate=True, override=None):
        if filename:
            self.filename = filename
        else:
            basename = 'iceprod_config.json'
            try:
                self.filename = locateconfig(basename)
            except Exception:
                logger.warning('config file does not exist, so making a new one')
                if 'I3PROD' in os.environ:
                    prefix = os.path.join(os.environ['I3PROD'], 'etc')
                else:
                    prefix = os.getcwd()
                self.filename = os.path.join(prefix, basename)

        self.validate = validate
        self.loading = False

        # load user input, apply defaults, and save
        self.load()
        if defaults:
            self.defaults()
        self.save()

        if override:
            self.apply_overrides(override)
            logger.info('after overrides: %s',self)

    def apply_overrides(self, overrides):
        for item in overrides:
            key,val = item.split('=',1)

            # try decoding value
            if val == 'true':
                val = True
            elif val == 'false':
                val = False
            elif val.isdigit():
                val = int(val)
            else:
                try:
                    val = float(val)
                except ValueError:
                    try:
                        val = json_decode(val)
                    except Exception:
                        pass

            # put value at right key
            key = key.split('.')
            logging.debug(f'setting {key}={val}')
            obj = self
            while len(key) > 1:
                if key[0] not in obj:
                    obj[key[0]] = {}
                obj = obj[key[0]]
                key = key[1:]
            obj[key[0]] = val

    def defaults(self):
        """Set default values if unset."""
        filename = None
        try:
            self.loading = True
            filename = get_pkgdata_filename('iceprod.server',
                                            'data/etc/config_defaults.json')
            text = open(filename).read()
            obj = json_decode(text)

            def setter(new_obj,self_obj):
                logger.debug('setter()')
                orig_keys = self_obj.keys()
                for key in new_obj:
                    logger.debug('key = %s',key)
                    if key == '*':
                        for key2 in orig_keys:
                            logger.debug('key2=%s',key2)
                            if isinstance(self_obj[key2],dict):
                                setter(new_obj['*'],self_obj[key2])
                    elif key not in self_obj:
                        logger.debug('setting key')
                        self_obj[key] = new_obj[key]
                    elif isinstance(self_obj[key],dict):
                        setter(new_obj[key],self_obj[key])
                logger.debug('~setter()')
            logger.info('before defaults: %s',self)
            setter(obj,self)
            # special case for site_id
            if 'site_id' not in self:
                self['site_id'] = GlobalID.siteID_gen()
                logger.warning('Generating new site_id: %s',self['site_id'])
            logger.info('with defaults: %s',self)
        except Exception:
            logger.warning('failed to load from default config file %r',
                           filename, exc_info=True)
        finally:
            self.loading = False

    def do_validate(self):
        if validate and self.validate:
            try:
                filename = get_pkgdata_filename('iceprod.server', 'data/etc/iceprod_schema.json')
                schema = json.load(open(filename))
                validate(self, schema)
            except ValidationError as e:
                path = '.'.join(e.path)
                logger.warning('Validation error at "%s": %s' % (path, e.message))
                raise e
        else:
            logger.warning('skipping validation of config')

    def load(self):
        """Load config from file, overwriting current contents."""
        try:
            self.loading = True
            if os.path.exists(self.filename):
                text = open(self.filename).read()
                obj = json_decode(text)
                for key in obj:
                    self[key] = obj[key]
                self.do_validate()
        except ValidationError:
            raise
        except Exception:
            logger.warning('failed to load from config file %s',self.filename,
                           exc_info=True)
        finally:
            self.loading = False

    def load_string(self, text):
        """Load a config from a string, saving to file."""
        try:
            self.loading = True
            obj = json_decode(text)
            for key in obj:
                self[key] = obj[key]
            self.do_validate()
        finally:
            self.loading = False
            self.save()

    def save_to_string(self):
        return json_encode(self, indent=4)

    def save(self):
        """Save config from file."""
        if not self.loading:
            try:
                text = json_encode(self, indent=4)
                # save securely
                with os.fdopen(os.open(self.filename+'.tmp', os.O_WRONLY | os.O_CREAT, 0o600),'w') as f:
                    f.write(text)
                os.rename(self.filename+'.tmp',self.filename)
            except Exception:
                logger.warning('failed to save to config file %s',self.filename,
                               exc_info=True)

    # insert save function into dict methods
    def __setitem__(self, key, value):
        super(IceProdConfig,self).__setitem__(key, value)
        self.save()

    def __delitem__(self, key):
        super(IceProdConfig,self).__delitem__(key)
        self.save()



"""
The Condor plugin.  Allows submission to
`HTCondor <http://research.cs.wisc.edu/htcondor/>`_.

Note: Condor was renamed to HTCondor in 2012.
"""
import os
import logging
import getpass
import subprocess
from functools import partial

from tornado.concurrent import run_on_executor

from iceprod.server import grid

logger = logging.getLogger('condor')


def check_call_clean_env(*args, **kwargs):
    env = os.environ.copy()
    del env['LD_LIBRARY_PATH']
    kwargs['env'] = env
    return subprocess.check_call(*args, **kwargs)


def check_output_clean_env(*args, **kwargs):
    env = os.environ.copy()
    del env['LD_LIBRARY_PATH']
    kwargs['env'] = env
    return subprocess.check_output(*args, **kwargs)


def condor_os_reqs(os_arch):
    """Convert from OS_ARCH to Condor OS requirements"""
    os_arch = os_arch.rsplit('_',2)[0].rsplit('.',1)[0]
    reqs = 'OpSysAndVer =?= "{}"'.format(os_arch.replace('RHEL','CentOS').replace('_',''))
    reqs = reqs + '|| OpSysAndVer =?= "{}"'.format(os_arch.replace('RHEL','SL').replace('_',''))
    reqs = reqs + ' || OSGVO_OS_STRING =?= "{}"'.format(os_arch.replace('_',' '))
    return '('+reqs+')'


class condor(grid.BaseGrid):
    """Plugin Overrides for HTCondor pilot submission"""

    # let the basic plugin be dumb and implement as little as possible

    @run_on_executor
    def generate_submit_file(self, task, cfg=None, passkey=None,
                             filelist=None):
        """Generate queueing system submit file for task in dir."""
        args = self.get_submit_args(task,cfg=cfg,passkey=passkey)

        # get requirements and batchopts
        requirements = []
        batch_opts = {}
        for b in self.queue_cfg['batchopts']:
            if b.lower() == 'requirements':
                requirements.append(self.queue_cfg['batchopts'][b])
            else:
                batch_opts[b] = self.queue_cfg['batchopts'][b]
        if cfg:
            if (cfg['steering'] and 'batchsys' in cfg['steering'] and
                    cfg['steering']['batchsys']):
                for b in cfg['steering']['batchsys']:
                    if b.lower().startswith(self.__class__.__name__):
                        # these settings apply to this batchsys
                        for bb in cfg['steering']['batchsys'][b]:
                            value = cfg['steering']['batchsys'][b][bb]
                            if bb.lower() == 'requirements':
                                requirements.append(value)
                            else:
                                batch_opts[bb] = value
            if task['task_id'] != 'pilot':
                if 'task' in cfg['options']:
                    t = cfg['options']['task']
                    if t in cfg['tasks']:
                        alltasks = [cfg['tasks'][t]]
                    else:
                        alltasks = []
                        try:
                            for tt in cfg['tasks']:
                                if t == tt['name']:
                                    alltasks.append(tt)
                        except Exception:
                            logger.warning('error finding specified task to run for %r',
                                           task,exc_info=True)
                else:
                    alltasks = cfg['tasks']
                for t in alltasks:
                    if 'batchsys' in t and t['batchsys']:
                        for b in t['batchsys']:
                            if b.lower().startswith(self.__class__.__name__):
                                # these settings apply to this batchsys
                                for bb in t['batchsys'][b]:
                                    value = t['batchsys'][b][bb]
                                    if bb.lower() == 'requirements':
                                        requirements.append(value)
                                    else:
                                        batch_opts[bb] = value

        # write the submit file
        submit_file = os.path.join(task['submit_dir'],'condor.submit')
        with open(submit_file,'w') as f:
            p = partial(print,sep='',file=f)
            p('universe = vanilla')
            p('executable = {}'.format(os.path.join(task['submit_dir'],'loader.sh')))
            p('log = condor.log')
            p('output = condor.out.$(Process)')
            p('error = condor.err.$(Process)')
            p('notification = never')
            p('+IsIceProdJob = True')  # mark as IceProd for monitoring
            p('want_graceful_removal = True')
            if filelist:
                p('transfer_input_files = {}'.format(','.join(filelist)))
                p('skip_filechecks = True')
                p('should_transfer_files = always')
                p('when_to_transfer_output = ON_EXIT_OR_EVICT')
                p('+SpoolOnEvict = False')
            p('transfer_output_files = iceprod_log, iceprod_out, iceprod_err')
            if 'num' in task:
                p('transfer_output_remaps = "iceprod_log=iceprod_log_$(Process)'
                  ';iceprod_out=iceprod_out_$(Process)'
                  ';iceprod_err=iceprod_err_$(Process)"')

            # handle resources
            p('+JobIsRunning = (JobStatus =!= 1) && (JobStatus =!= 5)')
            if 'reqs' in task:
                if 'cpu' in task['reqs'] and task['reqs']['cpu']:
                    p('+OriginalCpus = {}'.format(task['reqs']['cpu']))
                    p('+RequestResizedCpus = ((Cpus < OriginalCpus) ? OriginalCpus : Cpus)')
                    p('+JOB_GLIDEIN_Cpus = "$$(Cpus:0)"')
                    p('+JobIsRunningCpus = (JobIsRunning && (!isUndefined(MATCH_EXP_JOB_GLIDEIN_Cpus)))')
                    p('+JobCpus = (JobIsRunningCpus ? int(MATCH_EXP_JOB_GLIDEIN_Cpus) : OriginalCpus)')
                    p('request_cpus = (!isUndefined(Cpus)) ? RequestResizedCpus : JobCpus')
                    p('Rank = Rank + (isUndefined(Cpus) ? 0 : Cpus)/8')
                if 'gpu' in task['reqs'] and task['reqs']['gpu']:
                    p('+OriginalGpus = {}'.format(task['reqs']['gpu']))
                    p('+RequestResizedGpus = (Gpus < OriginalGpus) ? OriginalGpus : Gpus')
                    p('+JOB_GLIDEIN_Gpus = "$$(Gpus:0)"')
                    p('+JobIsRunningGpus = (JobIsRunning && (!isUndefined(MATCH_EXP_JOB_GLIDEIN_Gpus)))')
                    p('+JobGpus = (JobIsRunningGpus ? int(MATCH_EXP_JOB_GLIDEIN_GPUs) : OriginalGpus)')
                    p('request_gpus = !isUndefined(Gpus) ? RequestResizedGpus : JobGpus')
                if 'memory' in task['reqs'] and task['reqs']['memory']:
                    # extra 100MB for pilot
                    p('+OriginalMemory = {}'.format(int(task['reqs']['memory']*1000+100)))
                    p('+RequestResizedMemory = (Memory < OriginalMemory) ? OriginalMemory : Memory')
                    p('+JOB_GLIDEIN_Memory = "$$(Memory:0)"')
                    p('+JobIsRunningMemory = (JobIsRunning && (!isUndefined(MATCH_EXP_JOB_GLIDEIN_Memory)))')
                    p('+JobMemory = (JobIsRunningMemory ? int(MATCH_EXP_JOB_GLIDEIN_Memory) : OriginalMemory)')
                    p('request_memory = !isUndefined(Memory) ? RequestResizedMemory : JobMemory')
                else:
                    p('request_memory = 1000')
                if 'disk' in task['reqs'] and task['reqs']['disk']:
                    p('+OriginalDisk = {}'.format(int(task['reqs']['disk']*1000000)))
                    p('+RequestResizedDisk = (Disk-10000 < OriginalDisk) ? OriginalDisk : Disk-10000')
                    p('+JOB_GLIDEIN_Disk = "$$(Disk:0)"')
                    p('+JobIsRunningDisk = (JobIsRunning && (!isUndefined(MATCH_EXP_JOB_GLIDEIN_Disk)))')
                    p('+JobDisk = (JobIsRunningDisk ? int(MATCH_EXP_JOB_GLIDEIN_Disk) : OriginalDisk)')
                    p('request_disk = !isUndefined(Disk) ? RequestResizedDisk : JobDisk')
                if 'time' in task['reqs'] and task['reqs']['time']:
                    # extra 10 min for pilot
                    p('+OriginalTime = {}'.format(int(task['reqs']['time'])*3600+600))
                    p('+TargetTime = (!isUndefined(Target.PYGLIDEIN_TIME_TO_LIVE) ? Target.PYGLIDEIN_TIME_TO_LIVE : Target.TimeToLive)')
                    p('Rank = Rank + (TargetTime - OriginalTime)/86400')
                    requirements.append('TargetTime > OriginalTime')
                if 'os' in task['reqs'] and task['reqs']['os']:
                    requirements.append(condor_os_reqs(task['reqs']['os']))

            for b in batch_opts:
                p(b+'='+batch_opts[b])
            if requirements:
                p('requirements = ('+')&&('.join(requirements)+')')

            if task['task_id'] == 'pilot' and 'pilot_ids' in task:
                for pilot_id in task['pilot_ids']:
                    p('arguments = ',' '.join(args + ['--pilot_id', pilot_id]))
                    p('queue')
            elif 'num' in task:
                p('arguments = ',' '.join(args))
                p('queue {}'.format(task['num']))
            else:
                p('arguments = ',' '.join(args))
                p('queue')

    @run_on_executor
    def submit(self,task):
        """Submit task to queueing system."""
        cmd = ['condor_submit','-terse','condor.submit']
        out = check_output_clean_env(cmd, cwd=task['submit_dir'], universal_newlines=True)
        grid_queue_id = []
        for line in out.split('\n'):
            # look for range
            parts = [p.strip() for p in line.split('-') if p.strip()]
            if len(parts) != 2:
                continue
            major = parts[0].split('.')[0]
            minor_1 = int(parts[0].split('.')[1])
            minor_2 = int(parts[1].split('.')[1])
            for i in range(minor_1, minor_2+1):
                grid_queue_id.append('{}.{}'.format(major,i))
        task['grid_queue_id'] = ','.join(grid_queue_id)

    @run_on_executor
    def get_grid_status(self):
        """Get all tasks running on the queue system.
           Returns {grid_queue_id:{status,submit_dir}}
        """
        ret = {}
        cmd = ['condor_q',getpass.getuser(),'-af:j','jobstatus','cmd']
        out = check_output_clean_env(cmd, universal_newlines=True)
        print('get_grid_status():',out)
        for line in out.split('\n'):
            if not line.strip():
                continue
            gid,status,cmd = line.split()
            if 'loader.sh' not in cmd:
                continue
            if status == '1':
                status = 'queued'
            elif status == '2':
                status = 'processing'
            elif status == '4':
                status = 'completed'
            elif status in ('3','5','6'):
                status = 'error'
            else:
                status = 'unknown'
            ret[gid] = {'status':status,'submit_dir':os.path.dirname(cmd)}
        return ret

    @run_on_executor
    def remove(self,tasks):
        """Remove tasks from queueing system."""
        if tasks:
            check_call_clean_env(['condor_rm']+list(tasks))



"""
The queue module is responsible for interacting with the local batch or
queueing system, putting tasks on the queue and removing them as necessary.
"""
import asyncio
import os
import logging
import socket

import certifi

import iceprod
import iceprod.server
from iceprod.server import module
from iceprod.server.globus import SiteGlobusProxy
import iceprod.core.functions


class StopException(Exception):
    pass


logger = logging.getLogger('modules_queue')


class queue(module.module):
    """
    Run the queue module, which queues jobs onto the local grid system(s).
    """

    def __init__(self,*args,**kwargs):
        # run default init
        super(queue,self).__init__(*args,**kwargs)

        self.proxy = None

        self.max_duration = 3600*12

    def start(self):
        """Start the queue"""
        super(queue,self).start()

        # set up x509 proxy
        proxy_kwargs = {}
        if 'gridftp_cfgfile' in self.cfg['queue']:
            proxy_kwargs['cfgfile'] = self.cfg['queue']['gridftp_cfgfile']
        self.proxy = SiteGlobusProxy(**proxy_kwargs)

        # set up job cacert
        use_ssl = 'system' in self.cfg and 'ssl' in self.cfg['system'] and self.cfg['system']['ssl']
        if (use_ssl and 'cert' in self.cfg['system']['ssl']):
            if 'I3PROD' in os.environ:
                remote_cacert = os.path.expandvars(os.path.join('$I3PROD','etc','remote_cacert'))
            else:
                remote_cacert = os.path.expandvars(os.path.join('$PWD','remote_cacert'))
            with open(remote_cacert,'w') as f:
                f.write(open(certifi.where()).read())
                f.write('\n# IceProd local cert\n')
                f.write(open(self.cfg['system']['ssl']['cert']).read())
            self.cfg['system']['remote_cacert'] = remote_cacert

        # some setup
        self.plugins = []
        plugin_names = [x for x in self.cfg['queue'] if isinstance(self.cfg['queue'][x],dict)]
        plugin_cfg = [self.cfg['queue'][x] for x in plugin_names]
        plugin_types = [x['type'] for x in plugin_cfg]
        logger.info('queueing plugins in cfg: %r',{x:y for x,y in zip(plugin_names,plugin_types)})
        if not plugin_names:
            logger.debug('%r',self.cfg['queue'])
            logger.warning('no queueing plugins found. deactivating queue')
            self.stop()
            return

        # try to find plugins
        raw_types = iceprod.server.listmodules('iceprod.server.plugins')
        logger.info('available modules: %r',raw_types)
        plugins_tmp = []
        for i,t in enumerate(plugin_types):
            t = t.lower()
            p = None
            for r in raw_types:
                r_name = r.rsplit('.',1)[1].lower()
                if r_name == t:
                    # exact match
                    logger.debug('exact plugin match - %s',r)
                    p = r
                    break
                elif t.startswith(r_name):
                    # partial match
                    if p is None:
                        logger.debug('partial plugin match - %s',r)
                        p = r
                    else:
                        name2 = p.rsplit('.',1)[1]
                        if len(r_name) > len(name2):
                            logger.debug('better plugin match - %s',r)
                            p = r
            if p is not None:
                plugins_tmp.append((p,plugin_names[i],plugin_cfg[i]))
            else:
                logger.error('Cannot find plugin for grid %s of type %s',plugin_names[i],t)

        # instantiate all plugins that are required
        gridspec_types = {}
        if 'max_task_queued_time' in self.cfg['queue']:
            self.max_duration += self.cfg['queue']['max_task_queued_time']
        if 'max_task_processing_time' in self.cfg['queue']:
            self.max_duration += self.cfg['queue']['max_task_processing_time']
        for p,p_name,p_cfg in plugins_tmp:
            logger.warning('queueing plugin found: %s = %s', p_name, p_cfg['type'])
            # try instantiating the plugin
            args = (self.cfg['site_id']+'.'+p_name, p_cfg, self.cfg,
                    self.modules, self.executor, self.statsd,
                    self.rest_client, self.cred_client)
            try:
                self.plugins.append(iceprod.server.run_module(p,*args))
            except Exception:
                logger.error('Error importing plugin', exc_info=True)
            else:
                desc = p_cfg['description'] if 'description' in p_cfg else ''
                gridspec_types[self.cfg['site_id']+'.'+p_name] = {
                    'type': p_cfg['type'],
                    'description': desc,
                }
                duration = 0
                if 'max_task_queued_time' in p_cfg:
                    duration += p_cfg['max_task_queued_time']
                if 'max_task_processing_time' in p_cfg:
                    duration += p_cfg['max_task_processing_time']
                if duration > self.max_duration:
                    self.max_duration = duration

        # add gridspec and types to the db
        args = {
            'host': socket.getfqdn(),
            'queues': {p_name:p_cfg['type'] for p,p_name,p_cfg in plugins_tmp},
            'version': iceprod.__version__,
        }
        if 'grid_id' in self.cfg and self.cfg['grid_id']:
            try:
                self.rest_client.request_seq('GET', '/grids/{}'.format(self.cfg['grid_id']))
            except Exception:
                logger.warning('grid_id %s not present in DB',
                               self.cfg['grid_id'], exc_info=True)
                del self.cfg['grid_id']
        if 'grid_id' not in self.cfg:
            # register grid
            try:
                ret = self.rest_client.request_seq('POST', '/grids', args)
                self.cfg['grid_id'] = ret['result']
            except Exception:
                logger.fatal('cannot register grid in DB', exc_info=True)
                raise
        else:
            # update grid
            try:
                ret = self.rest_client.request_seq('PATCH', '/grids/{}'.format(self.cfg['grid_id']), args)
            except Exception:
                logger.warning('error updating grid in DB', exc_info=True)

        asyncio.create_task(self.queue_loop())

    async def queue_loop(self, run_once=False):
        """Run the queueing loop"""
        while True:
            # check and clean grids
            for p in self.plugins:
                try:
                    await p.check_and_clean()
                except Exception:
                    logger.error('plugin %s.check_and_clean() raised exception',
                                 p.__class__.__name__,exc_info=True)

            # check proxy cert
            try:
                self.check_proxy(self.max_duration)
            except Exception:
                logger.error('error checking proxy',exc_info=True)

            # queue tasks to grids
            for p in self.plugins:
                try:
                    await p.queue()
                except Exception:
                    logger.error('plugin %s.queue() raised exception',
                                 p.__class__.__name__,exc_info=True)

            if run_once:
                break

            # set timeout
            if 'queue' in self.cfg and 'queue_interval' in self.cfg['queue']:
                timeout = self.cfg['queue']['queue_interval']
                if timeout <= 0:
                    timeout = 300
            else:
                timeout = 300
            await asyncio.sleep(timeout)

    def check_proxy(self, duration=None):
        """
        Check the x509 proxy.

        Blocking function.
        """
        try:
            if duration:
                self.proxy.set_duration(duration//3600)
            self.proxy.update_proxy()
            self.cfg['queue']['x509proxy'] = self.proxy.get_proxy()
        except Exception:
            logger.warning('cannot setup x509 proxy', exc_info=True)
